#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>

#include "shfs.h"
#include "shfs_fcache.h"
#include "shfs_proc.h"

#define file_cache info->file_cache

void
shfs_fcache_init(struct shfs_sb_info *info)
{
	memset(&file_cache, 0, sizeof(struct shfs_file_cache));
}

int
shfs_fcache_empty(struct shfs_sb_info *info)
{
	int i;

	DEBUG("closing file cache...\n");
	for (i = 0; i < SHFS_FCACHE_HASH; i++) {
		if (file_cache.hash[i])
			VERBOSE("Aieeee.. file cache not empty!\n");
	}
	DEBUG(" OK\n");
	return 0;
}

static inline unsigned long 
shfs_file_cache_hash(unsigned long i)
{
	return i % SHFS_FCACHE_HASH;
}

int 
shfs_fcache_add(struct file *f)
{
	struct inode *inode;
        struct super_block *sb;
        struct shfs_sb_info *info; 
	unsigned long hsh;
	struct shfs_file_hash_node *p;

	DEBUG("Adding file to rw-cache.\n");

	if (f && f->f_dentry && (inode = f->f_dentry->d_inode)) {
		sb = inode->i_sb;
		DEBUG("ino: %lu\n", inode->i_ino);
		info = (struct shfs_sb_info*)sb->u.generic_sbp;
		hsh = shfs_file_cache_hash(inode->i_ino);
		for (p = file_cache.hash[hsh]; p; p = p->next) {
			if ((p->buf.ino == inode->i_ino)
			    && (p->buf.name_hash == f->f_dentry->d_name.hash))
				break;
		}
		if (p) {
			DEBUG("opened: %d times before\n", p->buf.used);
			p->buf.used++;
			if (!p->buf.used)
				VERBOSE("File usage counter overflow!\n");
		} else {
			p = (struct shfs_file_hash_node*)KMEM_ALLOC("fcache", file_hash_cache, GFP_KERNEL);
			if (!p) {
				VERBOSE("slab cache error!\n");
				return -1;
			}
			p->buf.data = vmalloc(SHFS_FCACHE_SIZE);
			if (!p->buf.data) {
				VERBOSE("slab cache error!\n");
				KMEM_FREE("fcache", file_hash_cache, p);
				return -1;
			}
			p->buf.ino = inode->i_ino;
			p->buf.used = 1;
			p->buf.name_hash = f->f_dentry->d_name.hash;
			p->buf.offset = 0;
			p->buf.count = 0;
			p->buf.type = SHFS_FCACHE_READ;
			DEBUG("ino: %lu (hash: %u)\n", p->buf.ino, p->buf.name_hash);
		
			hsh = shfs_file_cache_hash(inode->i_ino);
			p->prev = NULL;
			p->next = file_cache.hash[hsh];
			if (p->next)
				p->next->prev = p;
			file_cache.hash[hsh] = p;
			file_cache.len[hsh]++;
		}
		return 0;
	} 
	VERBOSE("Invalid file!\n");
	return -1;
}


int 
shfs_fcache_remove(struct file *f)
{
	struct inode *inode;
        struct super_block *sb;
        struct shfs_sb_info *info; 
	unsigned long hsh;
	struct shfs_file_hash_node *p;

	DEBUG("Removing file from rw-cache.\n");
	
	if (f && f->f_dentry && (inode = f->f_dentry->d_inode)) {
		DEBUG("ino: %lu\n", inode->i_ino);
		sb = inode->i_sb;
		info = (struct shfs_sb_info*)sb->u.generic_sbp;
		hsh = shfs_file_cache_hash(inode->i_ino);
		for (p = file_cache.hash[hsh]; p; p = p->next) {
			if ((p->buf.ino == inode->i_ino)
			    && (p->buf.name_hash == f->f_dentry->d_name.hash))
				break;
		}
		if (!p) {
			VERBOSE("File not found!\n");
			return -1;
		}
		DEBUG("ino: %lu (hash: %u)\n", p->buf.ino, p->buf.name_hash);
		
		p->buf.used--;
		if (!p->buf.used) {
			DEBUG("used count 0, releasing..\n");
			if (p->buf.count && p->buf.type == SHFS_FCACHE_WRITE)
				shfs_proc_write(f->f_dentry, p->buf.offset, p->buf.count, p->buf.data);
			vfree(p->buf.data);

			if (p->prev)
				p->prev->next = p->next;
			else
				file_cache.hash[hsh] = p->next;
			if (p->next)
				p->next->prev = p->prev;
			KMEM_FREE("fcache", file_hash_cache, p);
			file_cache.len[hsh]--;
		}
		return 0;
	}
	VERBOSE("Invalid file!\n");
	return -1;
}

int
shfs_fcache_sync(struct file *f)
{
	struct inode *inode;
        struct super_block *sb;
        struct shfs_sb_info *info; 
	unsigned long hsh;
	struct shfs_file_hash_node *p;

	DEBUG("Sync file in rw-cache.\n");
	
	if (f && f->f_dentry && (inode = f->f_dentry->d_inode)) {
		DEBUG("ino: %lu\n", inode->i_ino);
		sb = inode->i_sb;
		info = (struct shfs_sb_info*)sb->u.generic_sbp;
		hsh = shfs_file_cache_hash(inode->i_ino);
		for (p = file_cache.hash[hsh]; p; p = p->next) {
			if ((p->buf.ino == inode->i_ino)
			    && (p->buf.name_hash == f->f_dentry->d_name.hash))
				break;
		}
		if (!p) {
			VERBOSE("File not found!\n");
			return -1;
		}
		DEBUG("ino: %lu (hash: %u)\n", p->buf.ino, p->buf.name_hash);
		
		if (p->buf.count && p->buf.type == SHFS_FCACHE_WRITE) {
			DEBUG("Syncing..\n");
			shfs_proc_write(f->f_dentry, p->buf.offset, p->buf.count, p->buf.data);
			p->buf.type = SHFS_FCACHE_READ;
			p->buf.count = 0;
		}
		return 0;
	}			
	VERBOSE("Invalid file!\n");
	return -1;
}

int 
shfs_fcache_read(struct dentry *d, unsigned offset, unsigned count, char *buffer)
{
	struct inode *inode = d->d_inode;
	struct super_block *sb = inode->i_sb;
	struct shfs_sb_info *info = (struct shfs_sb_info*)sb->u.generic_sbp;
	unsigned long hsh, ino;
	unsigned readahead, c, o, x, y, z, read = 0;
	struct shfs_file_hash_node *p;
	int res;

	DEBUG("Reading file: offset: %u, count: %u\n", offset, count);

	if (d->d_inode && d->d_inode->i_ino) {
		ino = d->d_inode->i_ino;
		DEBUG("ino: %lu\n", ino);
		hsh = shfs_file_cache_hash(ino);
		for (p = file_cache.hash[hsh]; p; p = p->next) {
			if ((p->buf.ino == ino)
			    && (p->buf.name_hash == d->d_name.hash))
				break;
		}
		if (!p) {
			VERBOSE("File not found!\n");
			return -1;
		}
		DEBUG("ino: %lu (hash: %u)\n", p->buf.ino, p->buf.name_hash);

		/* hit? */
		if (offset >= p->buf.offset && offset < (p->buf.offset + p->buf.count)) {
			o = offset - p->buf.offset;
			c = count > p->buf.count - o ? p->buf.count - o : count;
			memcpy(buffer, p->buf.data + o, c);
			buffer += c;
			offset += c;
			count -= c;
			read += c;
		}

		if (count && p->buf.type == SHFS_FCACHE_WRITE) {
			shfs_proc_write(d, p->buf.offset, p->buf.count, p->buf.data);
			p->buf.type = SHFS_FCACHE_READ;
			p->buf.offset = offset;
			p->buf.count = 0;
		}

		while (count) {
			o = offset & PAGE_MASK;
			x = offset - o;
			if (p->buf.offset + p->buf.count == offset)
				readahead = p->buf.count;
			else {
				p->buf.offset = o;
				p->buf.count = 0;
				readahead = 0;
			}
			if (readahead < (x + count))
				readahead = x + count;
			readahead = (readahead+PAGE_SIZE-1) & PAGE_MASK;
			if (readahead > SHFS_FCACHE_SIZE)
				readahead = SHFS_FCACHE_SIZE;
			if (o % readahead) {	 /* HD */
				y = o, z = readahead;
				while (y && z)
					if (y > z) y %= z; else z %= y;
				readahead = y > z ? y : z;
			}
			if (p->buf.count + readahead > SHFS_FCACHE_SIZE) {
				p->buf.offset = o;
				p->buf.count = 0;
			}
			DEBUG("rw cache read: offset: %u, size %u\n", o, readahead);
			res = shfs_proc_read(d, o, readahead, p->buf.data+p->buf.count);
			if (res < 0) {
				p->buf.count = 0;
				return res;
			}

			c = res - x;
			if (c > count)
				c = count;
			memcpy(buffer, p->buf.data + p->buf.count + x, c);
			buffer += c;
			offset += c;
			count -= c;
			read += c;
			p->buf.count += res;
		}
		return read;
	}
	VERBOSE("Invalid file!\n");
	return -1;
}

int 
shfs_fcache_write(struct dentry *d, unsigned offset, unsigned count, char *buffer)
{
	struct inode *inode = d->d_inode;
	struct super_block *sb = inode->i_sb;
	struct shfs_sb_info *info = (struct shfs_sb_info*)sb->u.generic_sbp;
	unsigned long hsh, ino;
	unsigned o, c, wrote = 0;
	struct shfs_file_hash_node *p;
	int res = 0;

	DEBUG("Writing to file: offset: %u, count: %u\n", offset, count);
	
	if (d->d_inode && d->d_inode->i_ino) {
		ino = d->d_inode->i_ino;
		DEBUG("ino: %lu\n", ino);
		hsh = shfs_file_cache_hash(ino);
		for (p = file_cache.hash[hsh]; p; p = p->next) {
			if ((p->buf.ino == ino)
			    && (p->buf.name_hash == d->d_name.hash))
				break;
		}
		if (!p) {
			VERBOSE("File not found!\n");
			return -1;
		}
		DEBUG("ino: %lu (hash: %u)\n", p->buf.ino, p->buf.name_hash);

		if (p->buf.type == SHFS_FCACHE_READ) {
			p->buf.offset = offset;
			p->buf.count = 0;
			p->buf.type = SHFS_FCACHE_WRITE;
		}

		DEBUG("1 cache: offset: %u, count: %u\n", p->buf.offset, p->buf.count);
		if (offset >= p->buf.offset && offset < (p->buf.offset + SHFS_FCACHE_SIZE)) {
			o = offset - p->buf.offset;
			c = count > SHFS_FCACHE_SIZE - o ? SHFS_FCACHE_SIZE - o : count;
			memcpy(p->buf.data + o, buffer, c);
			if (o + c > p->buf.count)
				p->buf.count = o + c;
			buffer += c;
			offset += c;
			count -= c;
			wrote += c;
		}
		DEBUG("2 cache: offset: %u, count: %u\n", p->buf.offset, p->buf.count);
		while (count) {
			if ((res = shfs_proc_write(d, p->buf.offset, p->buf.count, p->buf.data)) < 0)
				break;
			c = count > SHFS_FCACHE_SIZE ? SHFS_FCACHE_SIZE : count;
			memcpy(p->buf.data, buffer, c);
			p->buf.offset = offset;
			p->buf.count = c;
			buffer += c;
			offset += c;
			count -= c;
			wrote += c;
		}
		DEBUG("3 cache: offset: %u, count: %u\n", p->buf.offset, p->buf.count);

		return res >= 0 ? wrote : res;
	}
	VERBOSE("Invalid file!\n");
	return -1;
}


