#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pagemap.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>

#include "shfs.h"
#include "shfs_proc.h"

static int shfs_file_readpage(struct file*, struct page*);
static int shfs_file_writepage(struct page*);
static int shfs_file_preparewrite(struct file*, struct page*, unsigned, unsigned);
static int shfs_file_commitwrite(struct file*, struct page*, unsigned, unsigned);
static int shfs_file_permission(struct inode*, int);
static int shfs_file_open(struct inode*, struct file*);
static int shfs_file_flush(struct file*);
static int shfs_file_release(struct inode*, struct file*);

struct file_operations shfs_file_operations = {
	read:		generic_file_read,
	write:		generic_file_write,
	mmap:		generic_file_mmap,
	open:		shfs_file_open,
	flush:		shfs_file_flush,
	release:	shfs_file_release,
};

struct inode_operations shfs_file_inode_operations = {
	permission:	shfs_file_permission,
	setattr:	shfs_notify_change,
};

struct address_space_operations shfs_file_aops = {
	readpage:	shfs_file_readpage,
	writepage:	shfs_file_writepage,
	prepare_write:	shfs_file_preparewrite,
	commit_write:	shfs_file_commitwrite
};

static int
shfs_file_readpage(struct file *f, struct page *p)
{
	struct dentry *dentry = f->f_dentry;
	struct shfs_sb_info *info = (struct shfs_sb_info*)dentry->d_sb->u.generic_sbp;
	char *buffer;
	unsigned long offset, count;
	int result;
	
	get_page(p);

	buffer = page_address(p);
	offset = p->index << PAGE_CACHE_SHIFT;
	count = PAGE_SIZE;

	lock_kernel();
	do {
		DEBUG(" \n");
		if (!shfs_lock(info)) {
			VERBOSE("Interrupted\n");
			unlock_kernel();
			result = -EINTR;
			goto io_error;
			
		}
		if (!info->mnt.disable_fcache)
			result = shfs_fcache_read(dentry, offset, count, buffer);
		else
			result = shfs_proc_read(dentry, offset, count, buffer);
		shfs_unlock(info);
		if (result < 0) {
			VERBOSE(" IO error! (%d)\n", result);
			unlock_kernel();
			goto io_error;
		}
		count -= result;
		offset += result;
		buffer += result;
		dentry->d_inode->i_atime = CURRENT_TIME;
		if (!result)
			break;
	} while (count);

	unlock_kernel();
	memset(buffer, 0, count);
	flush_dcache_page(p);
	SetPageUptodate(p);
	result = 0;
io_error:
	UnlockPage(p);	
	put_page(p);
	return shfs_remove_sigpipe(result);
}

static int
shfs_file_writepage(struct page *p)
{
	VERBOSE(" This should't happen!!\n");
	return -EFAULT;
}

static int
shfs_file_preparewrite(struct file *f, struct page *p, unsigned offset, unsigned to)
{
	DEBUG("page: %p, offset: %u, to: %u\n", p, offset, to);
	kmap(p);
	return 0;
}

static int
shfs_file_commitwrite(struct file *f, struct page *p, unsigned offset, unsigned to)
{
	struct dentry *dentry = f->f_dentry;
	struct shfs_sb_info *info = (struct shfs_sb_info*)dentry->d_sb->u.generic_sbp;
	struct inode *inode = p->mapping->host;
	char *buffer = page_address(p) + offset;
	int written = 0, result;
	unsigned count = to - offset;
	
	DEBUG("page: %p, offset: %u, to: %u\n", p, offset, to);
	offset += p->index << PAGE_CACHE_SHIFT;

	lock_kernel();
	
	if (info->mnt.readonly) {
		result = -EROFS;
		goto error;
	}

	do {
		if (!shfs_lock(info)) {
			VERBOSE("Interrupted\n");
			result = -EINTR;
			goto error;
		}
		if (!info->mnt.disable_fcache)
			result = shfs_fcache_write(dentry, offset, count, buffer);
		else
			result = shfs_proc_write(dentry, offset, count, buffer);
		shfs_unlock(info);
		if (result < 0) {
			VERBOSE(" IO error! (%d)\n", result);
			goto error;
		}

		count -= result;
		offset += result;
		buffer += result;
		written += result;
		dentry->d_inode->i_mtime = dentry->d_inode->i_atime = CURRENT_TIME;
		if (!result)
			break;

	} while (count);

//	memset(buffer, 0, count);
	result = 0;
error:
	unlock_kernel();
	kunmap(p);

	DEBUG("offset: %u, i_size: %lld\n", offset, inode->i_size);
	if (offset > inode->i_size) {
		struct iattr attr = {
			ia_valid: ATTR_SIZE,
			ia_size: offset
		};
		inode->i_size = offset;
		mark_inode_dirty(inode);
		if (!shfs_lock(info)) {
			VERBOSE("Interrupted\n");
			return -EINTR;
		}
		shfs_dcache_update_fattr(dentry, &attr);
		shfs_unlock(info);
	}

	return shfs_remove_sigpipe(result);
}

static int
shfs_file_permission(struct inode *inode, int mask)
{
	int mode = inode->i_mode;

	mode >>= 6;
	if ((mode & 7 & mask) != mask)
		return -EACCES;
	return 0;
}

static int
shfs_file_open(struct inode *inode, struct file *f)
{
	struct dentry *dentry = f->f_dentry;
	struct shfs_sb_info *info = (struct shfs_sb_info*)dentry->d_sb->u.generic_sbp;
	int flags = f->f_flags & O_ACCMODE;
	int res;

	DEBUG(" %s\n", dentry->d_name.name);
	res = 0;
	if (!info->mnt.disable_fcache) {
//		lock_kernel();
		if (!shfs_lock(info)) {
			VERBOSE("Interrupted\n");
			return -EINTR;
		}
		if (flags == O_RDONLY || flags == O_RDWR)
			res = shfs_proc_read(f->f_dentry, 0, 0, NULL);
		if ((res >= 0 && (flags == O_WRONLY || flags == O_RDWR))) {
			res = info->mnt.readonly ? -EROFS : shfs_proc_write(f->f_dentry, 0, 0, NULL);
		}
		if (res >= 0)
			shfs_fcache_add(f);
		shfs_unlock(info);
//		unlock_kernel();
	}
	return res;
}

static int
shfs_file_flush(struct file *f)
{
	struct dentry *dentry = f->f_dentry;
	struct shfs_sb_info *info = (struct shfs_sb_info*)dentry->d_sb->u.generic_sbp;
	int res;

	DEBUG(" %s\n", dentry->d_name.name);
	res = 0;
	if (!info->mnt.disable_fcache) {
		if (!shfs_lock(info)) {
			VERBOSE("Interrupted\n");
			return -EINTR;
		}
		res = shfs_fcache_sync(f);
		shfs_unlock(info);
	}
	return shfs_remove_sigpipe(res);
}

static int
shfs_file_release(struct inode *inode, struct file *f)
{
	struct dentry *dentry = f->f_dentry;
	struct shfs_sb_info *info = (struct shfs_sb_info*)dentry->d_sb->u.generic_sbp;
	int res;
	
	DEBUG(" %s\n", dentry->d_name.name);
	res = 0;
	if (atomic_read(&dentry->d_count) == 1 && !info->mnt.disable_fcache) {
		/* no way to return error */
		shfs_locku(info);
		res = shfs_fcache_remove(f);
		shfs_dcache_invalidate(dentry->d_parent);
		shfs_unlock(info);
	}
	return shfs_remove_sigpipe(res);
}

